# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee.models.layers.activations.gelu import GELU, gelu
from towhee.models.layers.activations.hardsigmoid import HardSigmoid, hard_sigmoid
from towhee.models.layers.activations.hardswish import HardSwish, hard_swish
from towhee.models.layers.activations.mish import Mish, mish
from towhee.models.layers.activations.hardmish import HardMish, hard_mish
from towhee.models.layers.activations.prelu import PReLU
from towhee.models.layers.activations.sigmoid import Sigmoid, sigmoid
from towhee.models.layers.activations.swish import Swish, swish
from towhee.models.layers.activations.tanh import Tanh, tanh

